'use strict';
var util = require('util');
function create(rules) {
    return new Validator(rules);
}
exports.create = create;
var Validator = (function () {
    function Validator(rules) {
        this.namedRules = {
            'uuid': /^([a-z0-9]{8}-)([a-z0-9]{4}-){3}([a-z0-9]{12})$/,
            'email': /\S+@\S+/,
            'datetime': /^\d{4}-\d{2}-\d{2}\s\d{2}:\d{2}:\d{2}(\.\d+)?$/,
            'text': /^[\S\s]+$/,
            'number': /^\d+$/,
            'bool': /^t(rue)?|f(alse)?$/i,
            'abbr': /^[A-Z]+$/,
            'alphanum': /^[a-z0-9]+$/i,
            'gender': /^[mf]$/i,
            'zip': /^[0-9\s]+$/i,
            'phone': /^\+?[0-9\s]+$/,
            'countrycode': /^[A-Z]{2}$/
        };
        this.rules = {};
        this.error = {};
        if (util.isObject(rules)) {
            for (var name_1 in rules) {
                this.addRule(name_1, rules[name_1]);
            }
        }
    }
    Validator.prototype.validate = function (value, validator) {
        if (this._isEmpty(validator)) {
            return true;
        }
        var item = this._meta(validator);
        if (this._isEmpty(value)) {
            return !item.required;
        }
        switch (item.type) {
            case 'regexp':
                return validator.test(value);
            case 'lambda':
                return validator(value);
            default:
                var rule = this.rules[item.type] || this.namedRules[item.type];
                if (rule === undefined) {
                    throw new Error(util.format('Unsupported validation type: %s', item.type));
                }
                if (util.isRegExp(rule)) {
                    return rule.test(value);
                }
                if (util.isFunction(rule)) {
                    return rule(value);
                }
        }
    };
    Validator.prototype.validateAll = function (item) {
        this.error = {};
        var name, validator, required, value, pass = true;
        for (name in this.rules) {
            validator = this.rules[name];
            required = this._meta(validator).required;
            value = this._resolve(item, name, required);
            if (this.validate(value, validator) === false) {
                this.error[name] = (value === undefined)
                    ? "Missing property: " + name
                    : "Validation error: given \"" + value + "\", expected \"" + validator + "\"";
                pass = false;
            }
        }
        return pass;
    };
    Validator.prototype.hasErrors = function () {
        return Object.keys(this.error).length > 0;
    };
    Validator.prototype.getErrors = function () {
        return this.error;
    };
    Validator.prototype.addRule = function (name, validator) {
        if (!util.isString(name)) {
            throw new Error(util.format('Parameter "name" must be string, %s given', typeof name));
        }
        if (util.isNull(validator)) {
            return this;
        }
        if (!util.isRegExp(validator) && !util.isFunction(validator) && !util.isString(validator)) {
            throw new Error(util.format('Parameter "validator" must be either RegExp, Function or String (named validator), %s given', typeof validator));
        }
        var meta = this._meta(validator);
        if (util.isString(validator) && this.namedRules[meta.type] === undefined) {
            throw new Error("Parameter \"validator\": invalid named validator \"" + validator + "\"");
        }
        this.rules[name] = util.isString(validator)
            ? this._getNamedRule(meta)
            : validator;
        return this;
    };
    Validator.prototype.removeRule = function (name) {
        delete this.rules[name];
        return this;
    };
    Validator.prototype.hasRule = function (name) {
        return !!this.rules[name];
    };
    Validator.prototype._getNamedRule = function (item) {
        return item.type + (item.required ? '' : ':null');
    };
    Validator.prototype._meta = function (validator) {
        var item = {
            type: null,
            required: null
        };
        if (this._isEmpty(validator)) {
            return item;
        }
        if (util.isString(validator)) {
            var part = validator.split(':');
            item.type = part[0];
            item.required = /^null|optional$/i.test(part[1] || 'required')
                ? false
                : true;
            return item;
        }
        if (util.isRegExp(validator)) {
            item.type = 'regexp';
            item.required = !/\^\$/.test(validator.toString());
            return item;
        }
        if (util.isFunction(validator)) {
            item.type = 'lambda';
            item.required = true;
            return item;
        }
        throw new Error(util.format('Unsupported validator type: %s', typeof validator));
    };
    Validator.prototype._isEmpty = function (value) {
        if (util.isRegExp(value))
            return false;
        if (util.isNullOrUndefined(value))
            return true;
        if (util.isString(value) && /^$/.test(value))
            return true;
        if (util.isArray(value) && !value.length)
            return true;
        if (util.isObject(value) && !Object.keys(value).length)
            return true;
        return false;
    };
    Validator.prototype._resolve = function (source, path, need) {
        if (!util.isUndefined(source[path]))
            return source[path];
        var value = source;
        var parts = path.split('.');
        var part;
        var i;
        for (i in parts) {
            part = parts[i];
            if (util.isUndefined(value[part])) {
                if (need)
                    this.error[path] = util.format("No such property: " + part);
                return undefined;
            }
            value = value[part];
        }
        return value;
    };
    return Validator;
}());
exports.Validator = Validator;
//# sourceMappingURL=index.js.map